# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: telegram_proxy/tlg_proxy.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    Optional,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class Request(betterproto.Message):
    bot_id: str = betterproto.string_field(1)
    token: str = betterproto.string_field(2)
    chat_id: str = betterproto.string_field(3)
    inline_buttons: bool = betterproto.bool_field(4)
    data: str = betterproto.string_field(5)


@dataclass(eq=False, repr=False)
class Response(betterproto.Message):
    bot_id: str = betterproto.string_field(1)
    data: str = betterproto.string_field(2)


class TlgProxyStub(betterproto.ServiceStub):
    async def post_message(
        self,
        request: "Request",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "Response":
        return await self._unary_unary(
            "/tlg.TlgProxy/PostMessage",
            request,
            Response,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_update(
        self,
        request: "Request",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "Response":
        return await self._unary_unary(
            "/tlg.TlgProxy/GetUpdate",
            request,
            Response,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class TlgProxyBase(ServiceBase):

    async def post_message(self, request: "Request") -> "Response":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_update(self, request: "Request") -> "Response":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_post_message(
        self, stream: "grpclib.server.Stream[Request, Response]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.post_message(request)
        await stream.send_message(response)

    async def __rpc_get_update(
        self, stream: "grpclib.server.Stream[Request, Response]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_update(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/tlg.TlgProxy/PostMessage": grpclib.const.Handler(
                self.__rpc_post_message,
                grpclib.const.Cardinality.UNARY_UNARY,
                Request,
                Response,
            ),
            "/tlg.TlgProxy/GetUpdate": grpclib.const.Handler(
                self.__rpc_get_update,
                grpclib.const.Cardinality.UNARY_UNARY,
                Request,
                Response,
            ),
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains Telegram proxy client implementation with secure SSL context.
It provides high-level abstraction for communication with Telegram proxy service.

Generate ssl certificates
cd ~/.MartinBinance/keys

Proxy Service pair

openssl req -x509 -days 365 -newkey rsa:2048 -nodes -subj '/CN=localhost' --addext 'subjectAltName=IP:aaa.bbb.ccc.ddd'\
 -keyout tlg-proxy.key -out tlg-proxy.pem

Client pair

openssl req -x509 -days 365 -newkey rsa:2048 -nodes -subj '/CN=localhost' -keyout tlg-client.key -out tlg-client.pem
"""
__author__ = "Jerry Fedorenko"
__copyright__ = "Copyright © 2025 Jerry Fedorenko aka VM"
__license__ = "MIT"
__version__ = "3.0.36"
__maintainer__ = "Jerry Fedorenko"
__contact__ = "https://github.com/DogsTailFarmer"

import ssl
from pathlib import Path
import asyncio
from typing import Any

import ujson as json
import toml
import random
import logging.handlers

from sqlalchemy.util import await_fallback

import martin_binance.tlg as tlg
from martin_binance import LOG_FILE_TLG, CONFIG_FILE, CERT_DIR
from martin_binance.lib import tasks_manage, tasks_cancel

from exchanges_wrapper import Channel, exceptions
#
logger = logging.getLogger('tlg_client')
logger.setLevel(logging.INFO)
formatter = logging.Formatter(fmt="[%(asctime)s: %(levelname)s] %(message)s")
#
fh = logging.handlers.RotatingFileHandler(LOG_FILE_TLG, maxBytes=1000000, backupCount=10)
fh.setFormatter(formatter)
logger.addHandler(fh)
fh.setLevel(logging.DEBUG)
#
sh = logging.StreamHandler()
sh.setFormatter(formatter)
logger.addHandler(sh)
sh.setLevel(logging.INFO)
#
config = toml.load(str(CONFIG_FILE))['Telegram']
TLG_PROXY_HOST = config['tlg_proxy_host']
TLG_PROXY_PORT = config['tlg_proxy_port']
SERVER_CERT = Path(CERT_DIR, "tlg-proxy.pem")
CLIENT_CERT = Path(CERT_DIR, "tlg-client.pem")
CLIENT_KEY = Path(CERT_DIR, "tlg-client.key")


def create_secure_context(client_cert: Path, client_key: Path, *, trusted: Path) -> ssl.SSLContext:
    ctx = ssl.create_default_context(cafile=str(trusted))
    ctx.minimum_version = ssl.TLSVersion.TLSv1_2
    ctx.load_cert_chain(str(client_cert), str(client_key))
    ctx.set_ciphers('ECDHE+AESGCM:ECDHE+CHACHA20:DHE+AESGCM:DHE+CHACHA20')
    ctx.set_alpn_protocols(['h2'])
    return ctx


SSL_CONTEXT = create_secure_context(CLIENT_CERT, CLIENT_KEY, trusted=SERVER_CERT)


class TlgClient:
    def __init__(self, bot_id, token, chat_id):
        self.bot_id = bot_id
        self.token = token
        self.chat_id = chat_id
        #
        self.channel = None
        self.stub = None
        self.init_event = asyncio.Event()
        self.init_event.set()
        #
        self.tasks = set()

    async def connect(self):
        self.init_event.clear()
        delay = 0
        while True:
            try:
                if self.channel:
                    self.channel.close()
                self.channel = Channel(TLG_PROXY_HOST, TLG_PROXY_PORT, ssl=SSL_CONTEXT)
                self.stub = tlg.TlgProxyStub(self.channel)
                await self.post_message("Connected", reraise=True)
                self.init_event.set()
                break
            except ConnectionRefusedError:
                delay += random.randint(1, 15)  # NOSONAR python:S2245
                logger.warning(f"Try connecting to Telegram proxy, retrying in {delay} second... ")
                await asyncio.sleep(delay)
            except ssl.SSLCertVerificationError as e:
                logger.error(f"Connect to Telegram proxy server failed: {e}")
                break
            except Exception as e:
                logger.error(f"Connect to Telegram proxy server failed, check certificate expiration date first: {e}")
                break

    async def post_message(self, text, inline_buttons=False, reraise=False):
        try:
            res = await self.stub.post_message(
                tlg.Request(
                    bot_id=self.bot_id,
                    token=self.token,
                    chat_id=self.chat_id,
                    inline_buttons=inline_buttons,
                    data=f"{self.bot_id}. {text}"
                )
            )
            return res
        except (ConnectionRefusedError, ConnectionAbortedError, exceptions.StreamTerminatedError, TimeoutError):
            if self.init_event.is_set():
                tasks_manage(self.tasks, self.connect())
            elif reraise:
                raise
        except ssl.SSLCertVerificationError as e:
            logger.error(f"Post message to Telegram proxy failed: {e}")
        except KeyboardInterrupt:
            pass  # user interrupt

    async def get_update(self) -> Any | None:
        try:
            res = await self.stub.get_update(
                tlg.Request(
                    bot_id=self.bot_id,
                )
            )
            return json.loads(res.data) if res else None
        except (ConnectionRefusedError, exceptions.StreamTerminatedError):
            if self.init_event.is_set():
                tasks_manage(self.tasks, self.connect())
        except KeyboardInterrupt:
            pass  # user interrupt
        except ssl.SSLCertVerificationError as e:
            logger.error(f"Get update from Telegram proxy failed: {e}")

    async def close(self):
        self.channel.close()
        await tasks_cancel(self.tasks)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Optimization of Trading Strategy Parameters
"""
__author__ = "Jerry Fedorenko"
__copyright__ = "Copyright © 2021 Jerry Fedorenko aka VM"
__license__ = "MIT"
__version__ = "3.0.1"
__maintainer__ = "Jerry Fedorenko"
__contact__ = "https://github.com/DogsTailFarmer"

from pathlib import Path
import optuna
import inquirer
from inquirer.themes import GreenPassion
from martin_binance import BACKTEST_PATH
from martin_binance.backtest.optimizer import optimize

SKIP_LOG = True

vis = optuna.visualization
ii_params = []


def main():
    while 1:
        questions = [
                inquirer.List(
                    "path",
                    message="Select from saved: exchange_PAIR with the strategy you want to optimize",
                    choices=[f.name for f in BACKTEST_PATH.iterdir() if f.is_dir() and f.name.count('_') == 1],
                ),
                inquirer.List(
                    "mode",
                    message="New study session or analise from saved one",
                    choices=["New", "Analise saved study session", "Exit"],
                ),
                inquirer.Text(
                    "n_trials",
                    message="Enter number of cycles, from 2 to 1000",
                    ignore=lambda x: x["mode"] in ("Analise saved study session", "Exit"),
                    default='150',
                    validate=lambda _, c: 15 <= int(c) <= 1000,
                ),
            ]

        answers = inquirer.prompt(questions, theme=GreenPassion())

        study_name = answers.get('path')  # Unique identifier of the study
        storage_name = f"sqlite:///{Path(BACKTEST_PATH, study_name, 'study.db')}"

        if answers.get('mode') == 'New':
            Path(BACKTEST_PATH, study_name, 'study.db').unlink(missing_ok=True)
            try:
                strategy = next(Path(BACKTEST_PATH, study_name).glob("cli_*.py"))
            except StopIteration:
                raise UserWarning(f"Can't find cli_*.py in {Path(BACKTEST_PATH, study_name)}")

            study = optimize(
                study_name,
                strategy,
                int(answers.get('n_trials', '0')),
                storage_name,
                skip_log=SKIP_LOG,
                show_progress_bar=SKIP_LOG
            )
            print_study_result(study)
            print(f"Study instance saved to {storage_name} for later use")
        elif answers.get('mode') == 'Analise saved study session':
            # noinspection PyArgumentList
            study = optuna.load_study(study_name=study_name, storage=storage_name)

            print(f"Best value: {study.best_value}")
            print(f"Original value: {study.get_trials()[0].value}")

            questions = [
                inquirer.List(
                    "mode",
                    message="Make a choice",
                    choices=["Plot from saved", "Get parameters for specific trial", "Exit"],
                ),
                inquirer.Text(
                    "n_trial",
                    message="Enter the trial number",
                    ignore=lambda x: x["mode"] in ("Plot from saved", "Exit"),
                ),
            ]

            answers = inquirer.prompt(questions, theme=GreenPassion())
            if answers.get('mode') == 'Plot from saved':
                i_params = print_study_result(study)
                for index, p in enumerate(i_params.items()):
                    ii_params.append(p[0])
                    if index == 2:
                        break
                #
                try:
                    fig = vis.plot_optimization_history(study)
                    fig.show()
                    contour_plot = vis.plot_contour(study, params=ii_params)
                    contour_plot.show()
                    slice_plot = vis.plot_slice(study, params=ii_params)
                    slice_plot.show()
                except ImportError:
                    print("Can't find GUI, you can copy study instance to another environment for analyze it")
            elif answers.get('mode') == 'Get parameters for specific trial':
                trial = study.get_trials()[int(answers.get('n_trial', '0'))]
                print(f"number: {trial.number}")
                print(trial.state)
                print(f"value: {trial.value}")
                print(f"params: {trial.params}")
            else:
                break
        else:
            break


def print_study_result(study):
    print(f"Optimal parameters: {study.best_params} for get {study.best_value}")
    try:
        importance_params = optuna.importance.get_param_importances(study)
    except RuntimeError as e:
        importance_params = {}
        print(e)
    else:
        print("Evaluate parameter importance based on completed trials in the given study:")
        for p in importance_params.items():
            print(p)
    return importance_params


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Free trading system for Binance SPOT API
"""
__author__ = "Jerry Fedorenko"
__copyright__ = "Copyright © 2021 Jerry Fedorenko aka VM"
__license__ = "MIT"
__version__ = "3.0.36"
__maintainer__ = "Jerry Fedorenko"
__contact__ = "https://github.com/DogsTailFarmer"

from pathlib import Path
from shutil import copy

from exchanges_wrapper.definitions import Interval

HEARTBEAT = 2  # Sec
ORDER_TIMEOUT = HEARTBEAT * 15  # Sec
KLINES_INIT = [Interval.ONE_MINUTE, Interval.FIFTY_MINUTES, Interval.ONE_HOUR]
#
WORK_PATH = Path(Path.home(), ".MartinBinance")
CONFIG_PATH = Path(WORK_PATH, "config")
CONFIG_FILE = Path(CONFIG_PATH, "ms_cfg.toml")
DB_FILE = Path(WORK_PATH, "funds_rate.db")
LOG_PATH = Path(WORK_PATH, "log")
LAST_STATE_PATH = Path(WORK_PATH, "last_state")
BACKTEST_PATH = Path(WORK_PATH, "back_test")
TRIAL_PARAMS = Path(WORK_PATH, "trial_params.json")
CERT_DIR = Path(WORK_PATH, "keys")
LOG_FILE_TLG = Path(LOG_PATH, "tlg_proxy.log")
EQUAL_STR = "================================================================"


def init():
    if CONFIG_FILE.exists():
        print(f"Client config found at {CONFIG_FILE}")
    else:
        print("Can't find client config file! Creating it...")
        for path in [CONFIG_PATH, LOG_PATH, LAST_STATE_PATH, CERT_DIR]:
            path.mkdir(parents=True, exist_ok=True)

        templates = Path(Path(__file__).parent.absolute(), "templates")

        copy(Path(templates, "ms_cfg.toml"), CONFIG_FILE)

        files_to_copy = [
            "funds_rate.db",
            "trial_params.json",
            "cli_0_BTCUSDT.py",
            "cli_1_BTCUSDT.py",
            "cli_2_TESTBTCTESTUSDT.py",
            "cli_3_BTCUSDT.py"
        ]
        # skipcq: PYL-W0106
        [copy(Path(templates, file_name), Path(WORK_PATH, file_name)) for file_name in files_to_copy]

        print(f"Before the first run, set the parameters in {CONFIG_FILE}")
        raise SystemExit(1)


if __name__ == '__main__':
    init()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
####################################################################
# Restart trade sessions saved in /last_state
##################################################################
__author__ = "Jerry Fedorenko"
__copyright__ = "Copyright © 2021-2025 Jerry Fedorenko aka VM"
__license__ = "MIT"
__version__ = "3.0.36"
__maintainer__ = "Jerry Fedorenko"
__contact__ = 'https://github.com/DogsTailFarmer'
##################################################################

import libtmux
import time
from martin_binance import Path, WORK_PATH, LAST_STATE_PATH

server = libtmux.Server()

while not server.has_session("Trade") or not server.windows.filter(name='srv'):
    time.sleep(1)

time.sleep(1)
session = server.sessions.get(session_name="Trade")

for window in session.windows:
    if window.name == 'srv':
        window.active_pane.send_keys('exchanges-wrapper-srv', enter=True)
    else:
        last_state = Path(LAST_STATE_PATH, f"{window.name.replace('-', '_').replace('/', '')}.json")
        pair = Path(WORK_PATH, f"cli_{window.name.replace('-', '_').replace('/', '')}.py")
        if pair.exists() and last_state.exists():
            window.active_pane.send_keys(f"{pair} 1", enter=True)
    time.sleep(4)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Searches for optimal parameters for a strategy under given conditions
"""
__author__ = "Jerry Fedorenko"
__copyright__ = "Copyright © 2024 Jerry Fedorenko aka VM"
__license__ = "MIT"
__version__ = "3.0.5"
__maintainer__ = "Jerry Fedorenko"
__contact__ = "https://github.com/DogsTailFarmer"


import importlib.util as iu
import logging.handlers
import stat
import sys
from decimal import Decimal
from pathlib import Path

import optuna
import ujson as json

from martin_binance import LOG_PATH, TRIAL_PARAMS

OPTIMIZER = Path(__file__).absolute()
try:
    OPTIMIZER.chmod(OPTIMIZER.stat().st_mode | stat.S_IEXEC)
except PermissionError:
    pass  # if executed in Docker environment

PARAMS_FLOAT = ['KBB']
STRATEGY = None


# noinspection PyUnusedLocal
def notify_exception(*args):
    pass  # Supress message from sys.excepthook


def any2str(_x) -> str:
    return f"{_x:.6f}".rstrip('0').rstrip('.')


def try_trade(mbs, skip_log, **kwargs):
    for key, value in kwargs.items():
        setattr(mbs.ex, key, value if isinstance(value, int) or key in PARAMS_FLOAT else Decimal(f"{value}"))
    mbs.ex.MODE = 'S'
    mbs.ex.SAVE_DS = False
    mbs.ex.LOGGING = not skip_log
    global STRATEGY
    STRATEGY = mbs.trade(STRATEGY)
    return float(mbs.ex.SESSION_RESULT.get('profit', 0)) + float(mbs.ex.SESSION_RESULT.get('free', 0))


def optimize(study_name, cli, n_trials, storage_name=None, _prm_best=None, skip_log=True, show_progress_bar=False):
    sys.excepthook = notify_exception
    optuna.logging.set_verbosity(optuna.logging.WARNING)

    # Load parameter definitions from JSON file
    with open(TRIAL_PARAMS) as f:
        param_defs = json.load(f)

    spec = iu.spec_from_file_location("strategy", cli)
    mbs = iu.module_from_spec(spec)
    spec.loader.exec_module(mbs)

    def objective(_trial):
        params = {}
        for param_name, param_props in param_defs.items():
            if param_props['type'] == 'int':
                params[param_name] = _trial.suggest_int(
                    param_name, *param_props['range'], step=param_props.get('step', 1)
                )
            elif param_props['type'] == 'float':
                params[param_name] = _trial.suggest_float(
                    param_name, *param_props['range'], step=param_props.get('step', 0.1)
                )
        return try_trade(mbs, skip_log, **params)

    # noinspection PyArgumentList
    _study = optuna.create_study(study_name=study_name, storage=storage_name, direction="maximize")

    if _prm_best:
        logger.info(f"Previous best params: {_prm_best}")
        _study.enqueue_trial(_prm_best)

    _study.optimize(objective, n_trials=n_trials, gc_after_trial=True, show_progress_bar=show_progress_bar)
    return _study


if __name__ == "__main__":
    logger = logging.getLogger('logger_S')
    logger.level = logging.INFO
    formatter = logging.Formatter(fmt="[%(asctime)s: %(levelname)s] %(message)s")
    #
    fh = logging.handlers.RotatingFileHandler(Path(LOG_PATH, sys.argv[6]), maxBytes=500000, backupCount=5)
    fh.setFormatter(formatter)
    fh.setLevel(logging.INFO)
    logger.addHandler(fh)
    #
    prm_best = json.loads(sys.argv[5])
    logger.info(f"Previous best params: {prm_best}")
    try:
        study = optimize(
            sys.argv[1],
            sys.argv[2],
            int(sys.argv[3]),
            storage_name=sys.argv[4],
            _prm_best=prm_best
        )
    except KeyboardInterrupt:
        pass  # ignore
    except Exception as ex:
        logger.info(f"optimizer: {ex}")
    else:
        new_value = round(study.best_value, ndigits=6)
        bp = {k: int(any2str(v)) if isinstance(v, int) else float(any2str(v)) for k, v in study.best_params.items()}

        logger.info(f"Optimal parameters: {bp} for get {new_value}")
        if new_value:
            logger.info(f"Importance parameters: {optuna.importance.get_param_importances(study)}")

        _value = round(study.get_trials()[0].value, ndigits=6)

        if not prm_best or new_value > _value:
            bp |= {'new_value': any2str(new_value), '_value': any2str(_value)}
            print(json.dumps(bp))
        else:
            print(json.dumps({}))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This script sets up a Telegram proxy server with secure SSL context.
It configures logging, loads configuration from a TOML file, and establishes
retry policies for HTTP requests. The script also defines a function to create
a secure context for SSL connections and provides a method to generate keyboard
markup for Telegram bot commands.
"""
__author__ = "Jerry Fedorenko"
__copyright__ = "Copyright © 2025 Jerry Fedorenko aka VM"
__license__ = "MIT"
__version__ = "3.0.36"
__maintainer__ = "Jerry Fedorenko"
__contact__ = "https://github.com/DogsTailFarmer"

import ssl
from pathlib import Path
import socket
import asyncio
import logging.handlers
import requests
import toml
from requests.adapters import HTTPAdapter, Retry
import ujson as json

import martin_binance.tlg as tlg
from martin_binance import LOG_FILE_TLG, CONFIG_FILE, CERT_DIR
from martin_binance.lib import tasks_manage
from exchanges_wrapper import Server, graceful_exit
#
logger = logging.getLogger('tlg_proxy')
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter(fmt="[%(asctime)s: %(levelname)s] %(message)s")
#
fh = logging.handlers.RotatingFileHandler(LOG_FILE_TLG, maxBytes=1000000, backupCount=10)
fh.setFormatter(formatter)
logger.addHandler(fh)
fh.setLevel(logging.DEBUG)
#
sh = logging.StreamHandler()
sh.setFormatter(formatter)
logger.addHandler(sh)
sh.setLevel(logging.INFO)
#
CLIENT_CERT = Path(CERT_DIR, "tlg-client.pem")
SERVER_CERT = Path(CERT_DIR, "tlg-proxy.pem")
SERVER_KEY = Path(CERT_DIR, "tlg-proxy.key")

config = toml.load(str(CONFIG_FILE))['Telegram']
TLG_URL = config['tlg_url']
TLG_PROXY_HOST = config['tlg_proxy_host']
TLG_PROXY_PORT = config['tlg_proxy_port']
HEARTBEAT = config['heartbeat']

SESSION = requests.Session()
retries = Retry(total=50, backoff_factor=1, status_forcelist=[101, 104, 111, 502, 503, 504])
SESSION.mount('https://', HTTPAdapter(max_retries=retries))
COMMAND_COUNT = 5


def create_secure_context(server_cert: Path, server_key: Path, *, trusted: Path) -> ssl.SSLContext:
    ctx = ssl.create_default_context(
        ssl.Purpose.CLIENT_AUTH,
        cafile=str(trusted),
    )
    ctx.minimum_version = ssl.TLSVersion.TLSv1_2
    ctx.verify_mode = ssl.CERT_REQUIRED
    ctx.load_cert_chain(str(server_cert), str(server_key))
    ctx.set_ciphers('ECDHE+AESGCM:ECDHE+CHACHA20:DHE+AESGCM:DHE+CHACHA20')
    ctx.set_alpn_protocols(['h2'])
    return ctx


def get_keyboard_markup():
    return json.dumps({
        "inline_keyboard": [
            [
                {'text': 'status', 'callback_data': 'status_callback'},
                {'text': 'stop', 'callback_data': 'stop_callback'},
                {'text': 'end', 'callback_data': 'end_callback'},
                {'text': 'restart', 'callback_data': 'restart_callback'},
                {'text': 'exit', 'callback_data': 'exit_callback'},
            ]
        ]})


def requests_post(_method, _data, inline_buttons=False):
    if inline_buttons:
        keyboard = get_keyboard_markup()
        _data['reply_markup'] = keyboard
    _res = None
    try:
        _res = SESSION.post(_method, data=_data)
    except requests.exceptions.RetryError as _exc:
        logger.error(f"Telegram: {_exc}")
    except Exception as _exc:
        logger.error(f"Telegram: {_exc}")
    return _res


def parse_command(token, chat_id, update_inner):
    update_id = update_inner.get('update_id')
    message = update_inner.get('message')
    from_id = message.get('from').get('id')
    if from_id != int(chat_id):
        return None
    message_id = message.get('message_id')
    text_in = update_inner.get('message').get('text')
    try:
        reply_to_message = message.get('reply_to_message').get('text')
    except AttributeError:
        reply_to_message = None
        _text = "The command must be a response to any message from a specific strategy," \
                " use Reply + Menu combination"
        requests_post(f'{TLG_URL}{token}/sendMessage', _data={'chat_id': chat_id, 'text': _text})
    return {
        'update_id': update_id,
        'message_id': message_id,
        'text_in': text_in,
        'reply_to_message': reply_to_message
    }


def parse_query(token, chat_id, update_inner):
    update_id = update_inner.get('update_id')
    query = update_inner.get('callback_query')
    from_id = query.get('from').get('id')
    if from_id != int(chat_id):
        return None
    message = query.get('message')
    message_id = message.get('message_id')
    query_data = query.get('data')
    reply_to_message = message.get('text')
    command = None
    if query_data == 'status_callback':
        command = 'status'
    elif query_data == 'stop_callback':
        command = 'stop'
    elif query_data == 'end_callback':
        command = 'end'
    elif query_data == 'restart_callback':
        command = 'restart'
    elif query_data == 'exit_callback':
        command = 'exit'
    requests_post(
        f'{TLG_URL}{token}//answerCallbackQuery',
        {'callback_query_id': query.get('id')}
    )
    return {
        'update_id': update_id,
        'message_id': message_id,
        'text_in': command,
        'reply_to_message': reply_to_message
    }


def telegram_get(token, chat_id, offset=None) -> list:
    command_list = []
    _method = f'{TLG_URL}{token}/getUpdates'
    _res = requests_post(_method, _data={'chat_id': chat_id, 'offset': offset})
    if not _res or _res.status_code != 200:
        if _res and _res.status_code != 200:
            logger.error(_res)
        return command_list
    __result = _res.json().get('result')
    for result_in in __result:
        parsed = None
        if result_in.get('message') is not None:
            parsed = parse_command(token, chat_id, result_in)
        if result_in.get('callback_query') is not None:
            parsed = parse_query(token, chat_id, result_in)
        if parsed:
            command_list.append(parsed)
    return command_list


def process_update(token, chat_id, update_inner):
    logger.debug(f"process_update.update_inner: {update_inner}")
    reply = update_inner.get('reply_to_message')
    if not reply:
        return
    in_bot_id = reply.split('.')[0]
    if in_bot_id not in TlgProxy.bot_ids:
        return
    msg_in = str(update_inner['text_in']).lower().strip().replace('/', '')
    TlgProxy.command[in_bot_id] = msg_in
    post_text = f"{in_bot_id}. received '{msg_in}' command, ran to do"
    requests_post(
        f'{TLG_URL}{token}/sendMessage',
        _data={'chat_id': chat_id, 'text': post_text}
    )


def set_bot_commands(token):
    # Set command for Telegram bot
    _command = requests_post(f'{TLG_URL}{token}/getMyCommands', _data=None)
    if _command and _command.status_code == 200 and (not _command.json().get('result') or
                                                     len(_command.json().get('result')) < COMMAND_COUNT):
        _commands = {
            "commands": json.dumps([
                {"command": "status",
                 "description": "Get strategy status"},
                {"command": "stop",
                 "description": "Stop strategy after end of cycle, not for Reverse"},
                {"command": "end",
                 "description": "Stop strategy after executed TP order, in Direct and Reverse, all the same"},
                {"command": "restart",
                 "description": "Restart current pair with recovery"},
                {"command": "exit",
                 "description": "Exit from apps as Ctrl-C locally"}
            ])
        }
        res = requests_post(f'{TLG_URL}{token}/setMyCommands', _data=_commands)
        logger.info(f"Set or update command menu for Telegram bot: code: {res.status_code}, result: {res.json()},"
                    f" restart Telegram bot by /start command for update it")


async def poll_update(request: tlg.Request):
    offset_id = None
    logger.info(f"New bot started, on request: {request.bot_id}")
    while True:
        await asyncio.sleep(HEARTBEAT)
        # Get external command from Telegram bot
        updates = telegram_get(request.token, request.chat_id, offset_id)
        if not updates:
            continue
        offset_id = updates[-1].get('update_id') + 1
        for update in updates:
            process_update(request.token, request.chat_id, update)


class TlgProxy(tlg.TlgProxyBase):
    bot_tokens = set()
    bot_ids = set()
    tasks = set()
    command = {}

    async def post_message(self, request: tlg.Request) -> tlg.Response:
        TlgProxy.bot_ids.add(request.bot_id)
        if request.token not in self.bot_tokens:
            TlgProxy.bot_tokens.add(request.token)
            set_bot_commands(request.token)
            # Start polling task
            tasks_manage(self.tasks, poll_update(request), name='poll_update')
        # Post message
        res = requests_post(
            f'{TLG_URL}{request.token}/sendMessage',
            _data={'chat_id': request.chat_id, 'text': request.data},
            inline_buttons=request.inline_buttons
        )
        return tlg.Response(bot_id=request.bot_id, data=json.dumps(res.status_code if res else None))

    async def get_update(self, request: tlg.Request):
        return tlg.Response(bot_id=request.bot_id, data=json.dumps(TlgProxy.command.pop(request.bot_id, None)))


def is_port_in_use(host: str, port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


async def main(host=TLG_PROXY_HOST, port=TLG_PROXY_PORT):
    if is_port_in_use(host, port):
        raise SystemExit(f"gRPC Telegram proxy: local port {port} already used")

    server = Server([TlgProxy()])
    with graceful_exit([server]):
        await server.start(
            host,
            port,
            ssl=create_secure_context(SERVER_CERT, SERVER_KEY, trusted=CLIENT_CERT)
        )
        logger.info(f"Starting Telegram proxy service on {host}:{port}")
        await server.wait_closed()


if __name__ == '__main__':
    asyncio.run(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
####################################################################
# Config for prometheus_client
# See README.md for detail
####################################################################
__author__ = "Jerry Fedorenko"
__copyright__ = "Copyright © 2021 Jerry Fedorenko aka VM"
__license__ = "MIT"
__version__ = "3.0.36"
__maintainer__ = "Jerry Fedorenko"
__contact__ = 'https://github.com/DogsTailFarmer'

import os
import time
import sqlite3
import psutil
from requests import Session
import toml
import platform
from prometheus_client import start_http_server, Gauge
from concurrent.futures import ThreadPoolExecutor

from martin_binance import Path, CONFIG_FILE, DB_FILE
from exchanges_wrapper import CONFIG_FILE as SRV_CONFIG_FILE

# region Import parameters

if not CONFIG_FILE.exists():
    if platform.system() == 'Darwin':
        user = (lambda: os.environ["USERNAME"] if "C:" in os.getcwd() else os.environ["USER"])()
        WORK_PATH = Path("Users", user, ".margin")
    else:
        WORK_PATH = Path().resolve()
    CONFIG_FILE = Path(WORK_PATH, "ms_cfg.toml")
    SRV_CONFIG_FILE = Path(WORK_PATH, "exch_srv_cfg.toml")
    DB_FILE = Path(WORK_PATH, "funds_rate.db")

config = toml.load(str(CONFIG_FILE)).get('Exporter')
accounts = toml.load(str(SRV_CONFIG_FILE)).get('accounts')

names = {acc['name']: acc['exchange'] for acc in accounts}
# external port for prometheus
PORT = config.get('port')

# sec delay for .db polling
SLEEP_TIME_S = config.get('sleep_time_s')

# Server name
VPS_NAME = config.get('vps_name')

# CoinMarketCap
URL = config.get('url')
API = config.get('api')
request_delay = 60 / config.get('rate_limit')
#  endregion

GET_RATE = True  # For develop purpose can skip getting asset rate from coinmarketcap
CURRENCY_RATE_LAST_TIME = int(time.time())

try:
    SQL_CONN = sqlite3.connect(DB_FILE, check_same_thread=False, timeout=10)
except sqlite3.Error as error:
    SQL_CONN = None
    print("SQLite error:", error)

# region Metric declare
STATUS_ALARM = Gauge("margin_alarm", "1 when not order", ['exchange', 'pair', 'vps_name'])
REQUEST_DELAY_G = Gauge("request_delay_g", "request delay in sec", ['vps_name'])

SUM_F_PROFIT = Gauge("margin_f_profit", "first profit", ['exchange', 'pair', 'vps_name'])
SUM_S_PROFIT = Gauge("margin_s_profit", "second profit", ['exchange', 'pair', 'vps_name'])
LAST_RATE = Gauge("margin_last_rate", "pair last rate", ['exchange', 'pair', 'vps_name'])
LAST_RATE_USD = Gauge("margin_last_rate_usd", "last rate second coin to USD", ['exchange', 'pair', 'vps_name'])
SUM_PROFIT = Gauge("margin_sum_profit", "sum profit on last rate", ['exchange', 'pair', 'vps_name'])
SUM_PROFIT_USD = Gauge("margin_sum_profit_usd", "sum profit on last rate on USD", ['exchange', 'pair', 'vps_name'])

CYCLE_COUNT = Gauge("margin_cycle_count", "cycle count", ['exchange', 'pair', 'vps_name'])
BUY_COUNT = Gauge("margin_buy_count", "cycle buy count", ['exchange', 'pair', 'vps_name'])
SELL_COUNT = Gauge("margin_sell_count", "cycle sell count", ['exchange', 'pair', 'vps_name'])

BUY_TIME = Gauge("margin_buy_time", "cycle buy time", ['exchange', 'pair', 'vps_name'])
SELL_TIME = Gauge("margin_sell_time", "cycle sell time", ['exchange', 'pair', 'vps_name'])
BUY_INTEREST = Gauge("margin_buy_interest", "sum buy interest", ['exchange', 'pair', 'vps_name'])
SELL_INTEREST = Gauge("margin_sell_interest", "sum sell interest", ['exchange', 'pair', 'vps_name'])

F_BALANCE = Gauge("margin_f_balance", "first balance amount", ['exchange', 'pair', 'vps_name'])
S_BALANCE = Gauge("margin_s_balance", "second balance amount", ['exchange', 'pair', 'vps_name'])
TOTAL_BALANCE = Gauge("margin_balance", "total balance amount by last rate", ['exchange', 'pair', 'vps_name'])
BALANCE_USD = Gauge("margin_balance_usd", "balance amount in USD", ['name', 'exchange', 'currency', 'vps_name'])

# Cycle parameters
CYCLE_BUY = Gauge("margin_cycle_buy", "cycle buy", ['exchange', 'pair', 'vps_name'])
OVER_PRICE = Gauge("margin_over_price", "over price", ['exchange', 'pair', 'vps_name'])
F_DEPO = Gauge("margin_f_depo", "first depo", ['exchange', 'pair', 'vps_name'])
S_DEPO = Gauge("margin_s_depo", "second depo", ['exchange', 'pair', 'vps_name'])

# VPS control
VPS_CPU = Gauge("margin_vps_cpu", "average cpu load", ['vps_name'])
VPS_MEMORY = Gauge("margin_vps_memory", "average memory use in %", ['vps_name'])

SET_ACTIVE = Gauge("margin_set_active", "mark rows as loaded and reset some metrics", ['vps_name'])
SET_ACTIVE.labels(VPS_NAME).set_function(lambda: set_active())  # skipcq: PYL-W0108

''' Cycle parameters for future use
PRICE_SHIFT = Gauge("margin_price_shift", "price shift", ['exchange', 'pair'])
PROFIT = Gauge("margin_profit", "profit", ['exchange', 'pair'])
ORDER_Q = Gauge("margin_order_q", "order_q", ['exchange', 'pair'])
MARTIN = Gauge("margin_martin", "martin", ['exchange', 'pair'])
LINEAR_GRID_K = Gauge("margin_linear_grid_k", "linear_grid_k", ['exchange', 'pair'])
ADAPTIVE_TRADE_CONDITION = Gauge("margin_adaptive_trade_condition", "adaptive_trade_condition", ['exchange', 'pair'])
KB = Gauge("margin_kb", "bollinger band k bottom", ['exchange', 'pair'])
KT = Gauge("margin_kt", "bollinger band k top", ['exchange', 'pair'])
'''
# endregion


def set_active():
    if PREPARED:
        try:
            SQL_CONN.execute('UPDATE t_funds SET active = 1 WHERE active = 0')
            SQL_CONN.commit()
        except sqlite3.Error as ex:
            print(f"Update t_funds failed: {ex}")
            return 0.0
        with ThreadPoolExecutor(max_workers=2) as executor:
            executor.submit(reset)
    return 1.0


# noinspection PyProtectedMember
def reset():
    time.sleep(5)
    metrics = [BUY_TIME, BUY_INTEREST, SELL_TIME, SELL_INTEREST]
    for metric in metrics:
        # noinspection PyUnresolvedReferences
        label_names = metric._labelnames
        label_values = metric._metrics.keys()
        for label_value in label_values:
            labels = dict(zip(label_names, label_value))
            metric.labels(**labels).set(0.0)


def get_rate(_currency_rate) -> dict:
    global request_delay
    replace = {
        'UST': 'USDT',
        'IOT': 'MIOTA',
        'TESTUSDT': 'USDT',
        'TESTBTC': 'BTC'
    }
    headers = {'Accepts': 'application/json', 'X-CMC_PRO_API_KEY': API}
    session = Session()
    session.headers.update(headers)
    buffer_rate = {}

    for currency in _currency_rate:
        _currency = replace.get(currency, currency)
        price = buffer_rate.get(_currency)
        if price is None and GET_RATE:
            price = -1
            parameters = {'amount': 1, 'symbol': 'USD', 'convert': _currency}
            try:
                response = session.get(URL, params=parameters)
            except Exception as er:
                print(er)
            else:
                if response.status_code == 429:
                    time.sleep(61)
                    request_delay *= 1.5
                    try:
                        response = session.get(URL, params=parameters)
                    except Exception as er:
                        print(er)
                if response.status_code == 200:
                    data = response.json()
                    price = data['data'][0]['quote'][_currency]['price'] or -1
                    buffer_rate[_currency] = price
        _currency_rate[currency] = price if GET_RATE else 1
    return _currency_rate


def db_handler(sql_conn, _currency_rate, currency_rate_last_time):
    global PREPARED
    PREPARED = False
    global request_delay
    cursor = sql_conn.cursor()
    # Aggregate score for pair on exchange
    cursor.execute('SELECT tex.name, tf.id_exchange,\
                    tf.f_currency, tf.s_currency,\
                    count(*) as cycle_count,\
                    sum(f_profit) as sum_f_profit,\
                    sum(s_profit) as sum_s_profit\
                    FROM t_funds as tf LEFT JOIN t_exchange tex USING(id_exchange)\
                    GROUP BY tex.name, tf.id_exchange, tf.f_currency, tf.s_currency')
    records = cursor.fetchall()
    # Get assets
    cursor.execute('SELECT tex.id_exchange, tex.name, ta.currency, ta.value\
                    FROM t_asset as ta LEFT JOIN t_exchange tex USING(id_exchange)\
                    WHERE ta.value > 0')
    assets = cursor.fetchall()
    # Create dict of used currencies
    for asset in assets:
        _currency_rate.setdefault(asset[2])
    for row in records:
        _currency_rate.setdefault(str(row[2]))
        _currency_rate.setdefault(str(row[3]))
    # Get currency rate for all currency from CoinMarketCap in relation to USD
    time_for_refresh = time.time() - currency_rate_last_time > 86400
    if None in _currency_rate.values() or time_for_refresh:
        get_rate(_currency_rate)
        currency_rate_last_time = int(time.time())
        REQUEST_DELAY_G.labels(VPS_NAME).set(request_delay)
        if request_delay > 60:
            request_delay = 60 / config.get('rate_limit')
    #
    F_BALANCE.clear()
    S_BALANCE.clear()
    TOTAL_BALANCE.clear()
    BALANCE_USD.clear()
    CYCLE_BUY.clear()
    F_DEPO.clear()
    S_DEPO.clear()
    OVER_PRICE.clear()
    BUY_TIME.clear()
    BUY_INTEREST.clear()
    SELL_TIME.clear()
    SELL_INTEREST.clear()
    #
    for row in records:
        # print(f"row: {row}")
        exchange = str(row[0])
        id_exchange = int(row[1])
        f_currency = str(row[2])
        s_currency = str(row[3])
        pair = f"{f_currency}/{s_currency}"
        CYCLE_COUNT.labels(exchange, pair, VPS_NAME).set(int(row[4]))
        sum_f_profit = float(row[5])
        SUM_F_PROFIT.labels(exchange, pair, VPS_NAME).set(sum_f_profit)
        sum_s_profit = float(row[6])
        SUM_S_PROFIT.labels(exchange, pair, VPS_NAME).set(sum_s_profit)
        # Alarm
        cursor.execute('SELECT order_buy, order_sell\
                        FROM t_orders\
                        WHERE id_exchange=:id_exchange\
                        AND f_currency=:f_currency\
                        AND s_currency=:s_currency',
                       {'id_exchange': id_exchange, 'f_currency': f_currency, 's_currency': s_currency})
        status_alarm = cursor.fetchone()
        alarm = 0
        if status_alarm:
            order_buy = int(status_alarm[0])
            order_sell = int(status_alarm[1])
            alarm = 0 if order_buy and order_sell else 1
        STATUS_ALARM.labels(exchange, pair, VPS_NAME).set(alarm)
        # Last rate
        cursor.execute('SELECT rate\
                        FROM t_funds\
                        WHERE id_exchange=:id_exchange\
                        AND f_currency=:f_currency\
                        AND s_currency=:s_currency\
                        ORDER BY id DESC LIMIT 1',
                       {'id_exchange': id_exchange, 'f_currency': f_currency, 's_currency': s_currency})
        if last_rate_row := cursor.fetchone():
            last_rate = float(last_rate_row[0])
            LAST_RATE.labels(exchange, pair, VPS_NAME).set(last_rate)
        else:
            last_rate = 0.0
        # Sum profit
        sum_profit = sum_f_profit * last_rate + sum_s_profit
        SUM_PROFIT.labels(exchange, pair, VPS_NAME).set(sum_profit)
        # Convert sum profit to USD by last rate
        sum_profit_usd = -1
        if _currency_rate.get(s_currency):
            try:
                sum_profit_usd = sum_profit / _currency_rate[s_currency]
                LAST_RATE_USD.labels(exchange, pair, VPS_NAME).set(_currency_rate[s_currency])
            except ZeroDivisionError:
                sum_profit_usd = -1
        SUM_PROFIT_USD.labels(exchange, pair, VPS_NAME).set(sum_profit_usd)

        # Cycle count, calculated by each buy and sell cycle
        cursor.execute('SELECT count(*)\
                        FROM t_funds\
                        WHERE id_exchange=:id_exchange\
                        AND f_currency=:f_currency\
                        AND s_currency=:s_currency\
                        AND cycle_buy = 1',
                       {'id_exchange': id_exchange, 'f_currency': f_currency, 's_currency': s_currency})
        cycle_buy_row = cursor.fetchone()
        cycle_buy_count = int(cycle_buy_row[0]) if cycle_buy_row[0] else 0
        cursor.execute('SELECT count(*)\
                        FROM t_funds\
                        WHERE id_exchange=:id_exchange\
                        AND f_currency=:f_currency\
                        AND s_currency=:s_currency\
                        AND cycle_buy = 0',
                       {'id_exchange': id_exchange, 'f_currency': f_currency, 's_currency': s_currency})
        cycle_sell_row = cursor.fetchone()
        cycle_sell_count = int(cycle_sell_row[0]) if cycle_sell_row[0] else 0
        BUY_COUNT.labels(exchange, pair, VPS_NAME).set(cycle_buy_count)
        SELL_COUNT.labels(exchange, pair, VPS_NAME).set(cycle_sell_count)

        # Sum income interest, calculated by each buy and sell cycle
        cursor.execute('SELECT sum(100 * s_profit / s_depo), sum(cycle_time)\
                        FROM t_funds\
                        WHERE id_exchange=:id_exchange\
                        AND f_currency=:f_currency\
                        AND s_currency=:s_currency\
                        AND cycle_buy = 1\
                        AND active = 0',
                       {'id_exchange': id_exchange, 'f_currency': f_currency, 's_currency': s_currency})
        cycle_buy_row = cursor.fetchone()
        cycle_buy_interest = float(cycle_buy_row[0]) if cycle_buy_row[0] else 0.0
        cycle_buy_time = float(cycle_buy_row[1]) if cycle_buy_row[1] else 0.0

        cursor.execute('SELECT sum(100 * f_profit / f_depo), sum(cycle_time)\
                        FROM t_funds\
                        WHERE id_exchange=:id_exchange\
                        AND f_currency=:f_currency\
                        AND s_currency=:s_currency\
                        AND cycle_buy = 0\
                        AND active = 0',
                       {'id_exchange': id_exchange, 'f_currency': f_currency, 's_currency': s_currency})
        cycle_sell_row = cursor.fetchone()
        cycle_sell_interest = float(cycle_sell_row[0]) if cycle_sell_row[0] else 0.0
        cycle_sell_time = float(cycle_sell_row[1]) if cycle_sell_row[1] else 0.0

        BUY_TIME.labels(exchange, pair, VPS_NAME).set(cycle_buy_time)
        BUY_INTEREST.labels(exchange, pair, VPS_NAME).set(cycle_buy_interest)
        SELL_TIME.labels(exchange, pair, VPS_NAME).set(cycle_sell_time)
        SELL_INTEREST.labels(exchange, pair, VPS_NAME).set(cycle_sell_interest)

        # Balance amount
        cursor.execute('SELECT f_balance, s_balance, cycle_buy, f_depo, s_depo, over_price\
                        FROM t_funds\
                        WHERE id_exchange=:id_exchange\
                        AND f_currency=:f_currency\
                        AND s_currency=:s_currency\
                        ORDER BY id DESC LIMIT 1',
                       {'id_exchange': id_exchange, 'f_currency': f_currency, 's_currency': s_currency})
        if balance_row := cursor.fetchone():
            f_balance = balance_row[0]
            s_balance = balance_row[1]
            balance = f_balance * last_rate + s_balance
            F_BALANCE.labels(exchange, pair, VPS_NAME).set(f_balance)
            S_BALANCE.labels(exchange, pair, VPS_NAME).set(s_balance)
            TOTAL_BALANCE.labels(exchange, pair, VPS_NAME).set(balance)
            # Cycle parameters
            CYCLE_BUY.labels(exchange, pair, VPS_NAME).set(balance_row[2])
            F_DEPO.labels(exchange, pair, VPS_NAME).set(balance_row[3])
            S_DEPO.labels(exchange, pair, VPS_NAME).set(balance_row[4])
            OVER_PRICE.labels(exchange, pair, VPS_NAME).set(balance_row[5])

    for asset in assets:
        if _rate := _currency_rate.get(asset[2]):
            try:
                usd_amount = asset[3] / _rate
            except ZeroDivisionError:
                usd_amount = -1
            if usd_amount >= 1.0:
                BALANCE_USD.labels(names.get(asset[1]), asset[1], asset[2], VPS_NAME).set(usd_amount)

    cursor.close()
    return currency_rate_last_time


if __name__ == '__main__':
    # Start up the server to expose the metrics.
    currency_rate = {}
    start_http_server(PORT, addr='::')
    while True:
        try:
            CURRENCY_RATE_LAST_TIME = db_handler(SQL_CONN, currency_rate, CURRENCY_RATE_LAST_TIME)
        except sqlite3.Error as error:
            print("DB operational error:", error)
        else:
            PREPARED = True
        VPS_CPU.labels(VPS_NAME).set(100 * psutil.getloadavg()[0] / psutil.cpu_count())
        #
        memory = psutil.virtual_memory()
        swap = psutil.swap_memory()
        total_used_percent = 100 * float(swap.used + memory.used) / (swap.total + memory.total)
        VPS_MEMORY.labels(VPS_NAME).set(total_used_percent)
        #
        time.sleep(SLEEP_TIME_S)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Simple exchange simulator for backtest purpose
"""
__author__ = "Jerry Fedorenko"
__copyright__ = "Copyright © 2021 Jerry Fedorenko aka VM"
__license__ = "MIT"
__version__ = "3.0.34"
__maintainer__ = "Jerry Fedorenko"
__contact__ = "https://github.com/DogsTailFarmer"

from decimal import Decimal
from typing import Dict

import pandas as pd


def any2str(_x) -> str:
    return f"{_x:.8f}".rstrip('0').rstrip('.')


class Funds:
    __slots__ = ("base", "quote")

    def __init__(self):
        # {'asset': 'BTC', 'free': '0.0', 'locked': '0.0'}
        self.base = {}
        self.quote = {}

    def get_funds(self):
        base = self.base.copy()
        base |= {'free': str(base['free']), 'locked': str(base['locked'])}
        quote = self.quote.copy()
        quote |= {'free': str(quote['free']), 'locked': str(quote['locked'])}
        return [base, quote]

    def on_order_created(self, buy: bool, amount: Decimal, price: Decimal):
        if buy:
            self.quote['free'] -= amount * price
            self.quote['locked'] += amount * price
        else:
            self.base['free'] -= amount
            self.base['locked'] += amount

    def on_order_canceled(self, side: str, amount: Decimal, price: Decimal):
        if side == 'BUY':
            self.quote['free'] += amount * price
            self.quote['locked'] -= amount * price
        else:
            self.base['free'] += amount
            self.base['locked'] -= amount

    def on_order_filled(self, side: str, amount: Decimal, price: Decimal, last_price: Decimal, fee: Decimal):
        if side == 'BUY':
            self.base['free'] += amount - fee * amount / 100
            self.quote['locked'] -= amount * price
            self.quote['free'] += amount * (price - last_price)
        else:
            self.base['locked'] -= amount
            self.quote['free'] += amount * last_price - fee * (amount * last_price) / 100


class Order:
    __slots__ = (
        "symbol",
        "order_id",
        "order_list_id",
        "client_order_id",
        "transact_time",
        "price",
        "orig_qty",
        "executed_qty",
        "cummulative_quote_qty",
        "status",
        "time_in_force",
        "type",
        "side",
        "working_time",
        "self_trade_prevention_mode",
        "event_time",
        "last_executed_quantity",
        "cumulative_filled_quantity",
        "last_executed_price",
        "trade_id",
        "order_creation_time",
        "quote_asset_transacted",
        "last_quote_asset_transacted",
        "quote_order_quantity",
    )

    def __init__(self, symbol: str, order_id: int, client_order_id: str, buy: bool, amount: str, price: str, lt: int):
        self.symbol = symbol
        self.order_id = order_id
        self.order_list_id = -1
        self.client_order_id = client_order_id
        self.transact_time = lt  # local time
        self.price = Decimal(price)
        self.orig_qty = Decimal(amount)
        self.executed_qty = Decimal('0')
        self.cummulative_quote_qty = Decimal('0')
        self.status = 'NEW'
        self.time_in_force = 'GTC'
        self.type = 'LIMIT'
        self.side = "BUY" if buy else "SELL"
        self.working_time = "-1"
        self.self_trade_prevention_mode = 'NONE'
        #
        self.event_time: int
        self.last_executed_quantity = Decimal('0')
        self.cumulative_filled_quantity = Decimal('0')
        self.last_executed_price = Decimal('0')
        self.trade_id: int
        self.order_creation_time = lt
        self.quote_asset_transacted = Decimal('0')
        self.last_quote_asset_transacted = Decimal('0')
        self.quote_order_quantity = self.orig_qty * self.price


class Account:
    __slots__ = (
        "save_ds",
        "funds",
        "fee_maker",
        "fee_taker",
        "orders",
        "orders_buy",
        "orders_sell",
        "trade_id",
        "ticker",
        "grid_buy",
        "grid_sell",
        "ticker_last",
        "market_ids",
    )

    def __init__(self, save_ds: bool):
        self.save_ds = save_ds
        self.funds = Funds()
        self.fee_maker = Decimal('0')
        self.fee_taker = Decimal('0')
        self.orders = {}
        self.orders_buy = pd.Series()
        self.orders_sell = pd.Series()
        self.trade_id = 0
        self.ticker = {}
        self.grid_buy = {}
        self.grid_sell = {}
        self.ticker_last = Decimal('0')
        self.market_ids = []

    def create_order(
            self,
            symbol: str,
            client_order_id: str,
            buy: bool,
            amount: str,
            price: str,
            lt: int,
            order_id=None) -> {}:

        order_id = order_id or ((max(self.orders.keys()) + 1) if self.orders else 1)
        order = Order(
            symbol=symbol,
            order_id=order_id,
            client_order_id=client_order_id,
            buy=buy,
            amount=amount,
            price=price,
            lt=lt
        )

        if buy:
            self.orders_buy.at[order_id] = Decimal(price)
            if self.save_ds:
                self.grid_buy[lt] = self.orders_buy
        else:
            self.orders_sell.at[order_id] = Decimal(price)
            if self.save_ds:
                self.grid_sell[lt] = self.orders_sell
            #
        self.funds.on_order_created(buy=buy, amount=Decimal(amount), price=Decimal(price))
        self.orders[order_id] = order

        if self.ticker_last and ((buy and Decimal(price) >= self.ticker_last) or
                                 (not buy and Decimal(price) <= self.ticker_last)):
            # Market event
            self.market_ids.append(order_id)

        return {'symbol': order.symbol,
                'orderId': order.order_id,
                'orderListId': order.order_list_id,
                'clientOrderId': order.client_order_id,
                'transactTime': order.transact_time,
                'price': order.price,
                'origQty': order.orig_qty,
                'executedQty': order.executed_qty,
                'cummulativeQuoteQty': order.cummulative_quote_qty,
                'status': order.status,
                'timeInForce': order.time_in_force,
                'type': order.type,
                'side': order.side,
                'workingTime': order.working_time,
                'selfTradePreventionMode': order.self_trade_prevention_mode}

    def cancel_order(self, order_id: int, ts: int):
        order = self.orders.get(order_id)
        if order is None:
            raise UserWarning(f"Error on Cancel order, can't find {order_id} anymore")

        order.status = 'CANCELED'
        try:
            if order.side == 'BUY':
                self.orders_buy = self.orders_buy.drop(order_id)
                if self.save_ds and self.orders_buy.values.size:
                    self.grid_buy[ts] = self.orders_buy
            else:
                self.orders_sell = self.orders_sell.drop(order_id)
                if self.save_ds and self.orders_sell.values.size:
                    self.grid_sell[ts] = self.orders_sell
        except Exception as ex:
            raise UserWarning(f"Order {order_id} not active: {ex}") from ex

        self.orders[order_id] = order
        self.funds.on_order_canceled(order.side, order.orig_qty - order.executed_qty, order.price)
        return {'symbol': order.symbol,
                'origClientOrderId': order.client_order_id,
                'orderId': order.order_id,
                'orderListId': order.order_list_id,
                'clientOrderId': 'qwert',
                'price': str(order.price),
                'origQty': str(order.orig_qty),
                'executedQty': str(order.executed_qty),
                'cummulativeQuoteQty': str(order.cummulative_quote_qty),
                'status': order.status,
                'timeInForce': order.time_in_force,
                'type': order.type,
                'side': order.side,
                'selfTradePreventionMode': order.self_trade_prevention_mode}

    def on_ticker_update(self, ticker: {}, ts: int) -> list[Dict]:
        filled_buy_id = []
        filled_sell_id = []
        orders_id = []
        orders_filled = []

        self.ticker_last = Decimal(ticker['lastPrice'])
        qty = Decimal(ticker['Qty'])
        part = bool(qty)

        if self.market_ids:
            orders_id.extend(self.market_ids)

        orders_id.extend(self.orders_buy[self.orders_buy >= self.ticker_last].index.values)
        orders_id.extend(self.orders_sell[self.orders_sell <= self.ticker_last].index.values)

        if self.save_ds:
            # Save data for analytics
            self.ticker[ts] = ticker['lastPrice']
            if self.orders_sell.values.size:
                self.grid_sell[ts] = self.orders_sell
            if self.orders_buy.values.size:
                self.grid_buy[ts] = self.orders_buy
        #
        for order_id in orders_id:
            if part and not qty:
                break

            order = self.orders.get(order_id)

            order.transact_time = int(ticker['closeTime'])
            order.event_time = order.transact_time
            order.trade_id = self.trade_id = self.trade_id + 1

            order.last_executed_price = self.ticker_last

            delta = order.orig_qty - order.executed_qty
            order.last_executed_quantity = last_executed_qty = min(delta, qty) if part else delta
            order.executed_qty += last_executed_qty
            order.last_quote_asset_transacted = order.last_executed_price * last_executed_qty
            order.quote_asset_transacted += order.last_quote_asset_transacted

            if part:
                qty -= last_executed_qty

            order.cumulative_filled_quantity = order.executed_qty
            order.cummulative_quote_qty = order.quote_asset_transacted

            if order.executed_qty >= order.orig_qty:
                order.status = 'FILLED'
                if order.side == 'BUY':
                    filled_buy_id.append(order_id)
                else:
                    filled_sell_id.append(order_id)
            elif 0 < order.executed_qty < order.orig_qty:
                order.status = 'PARTIALLY_FILLED'
            #
            self.orders[order_id] = order
            #
            res = {
                'event_time': order.event_time,
                'symbol': order.symbol,
                'client_order_id': order.client_order_id,
                'side': order.side,
                'order_type': order.type,
                'time_in_force': order.time_in_force,
                'order_quantity': str(order.orig_qty),
                'order_price': str(order.price),
                'stop_price': '0',
                'iceberg_quantity': '0',
                'order_list_id': -1,
                'original_client_id': order.client_order_id,
                'execution_type': 'TRADE',
                'order_status': order.status,
                'order_reject_reason': 'NONE',
                'order_id': order_id,
                'last_executed_quantity': str(order.last_executed_quantity),
                'cumulative_filled_quantity': str(order.cumulative_filled_quantity),
                'last_executed_price': str(order.last_executed_price),
                'commission_amount': '0',
                'commission_asset': '',
                'transaction_time': order.transact_time,
                'trade_id': order.trade_id,
                'ignore_a': 12345678,
                'in_order_book': False,
                'is_maker_side': bool(order_id not in self.market_ids),
                'ignore_b': True,
                'order_creation_time': order.order_creation_time,
                'quote_asset_transacted': str(order.quote_asset_transacted),
                'last_quote_asset_transacted': str(order.last_quote_asset_transacted),
                'quote_order_quantity': str(order.quote_order_quantity)
            }
            #
            orders_filled.append(res)

            self.funds.on_order_filled(
                order.side,
                order.last_executed_quantity,
                order.price,
                order.last_executed_price,
                self.fee_taker if order_id in self.market_ids else self.fee_maker
            )
            #
        self.orders_buy = self.orders_buy.drop(filled_buy_id, errors="ignore")
        self.orders_sell = self.orders_sell.drop(filled_sell_id, errors="ignore")
        self.market_ids.clear()

        return orders_filled

    def restore_state(self, symbol: str, lt: int, orders: list, sum_amount: ()):
        if sum_amount[0]:
            self.funds.base['free'] += sum_amount[1]
            self.funds.quote['free'] -= sum_amount[2]
        else:
            self.funds.base['free'] -= sum_amount[1]
            self.funds.quote['free'] += sum_amount[2]

        for order in orders:
            self.create_order(
                symbol=symbol,
                client_order_id='',
                buy=order['buy'],
                amount=any2str(order['amount']),
                price=any2str(order['price']),
                lt=lt,
                order_id=order['id']
            )



"""
martin-binance classes and methods definitions
"""
__author__ = "Jerry Fedorenko"
__copyright__ = "Copyright © 2021 Jerry Fedorenko aka VM"
__license__ = "MIT"
__version__ = "3.0.36"
__maintainer__ = "Jerry Fedorenko"
__contact__ = "https://github.com/DogsTailFarmer"

import asyncio
import inspect
import logging.handlers
import time
from decimal import Decimal, ROUND_CEILING, ROUND_FLOOR, ROUND_HALF_EVEN
from enum import Enum
from pathlib import Path

import numpy as np
import ujson as json
from scipy.optimize import minimize

logger = logging.getLogger(f'logger.{__name__}')
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(logging.Formatter(fmt="[%(asctime)s: %(levelname)s] %(message)s"))
stream_handler.setLevel(logging.INFO)
logger.addHandler(stream_handler)

O_DEC = Decimal()


def tasks_manage(tasks_set: set, coro, name=None, add_done_callback=True):
    name = f"{name if name else ''}{'-' if name else ''}{coro.__name__}-{inspect.stack()[1][3]}"
    _t = asyncio.create_task(coro, name=name)
    tasks_set.add(_t)
    if add_done_callback:
        _t.add_done_callback(tasks_set.discard)


async def tasks_cancel(tasks_set: set, name=None, log_out=True):
    tasks = tasks_set.copy()
    for task in tasks:
        logger.debug(f"tasks_cancel.task_name: {task.get_name()}")
        if name and f"{name}" not in task.get_name():
            continue
        task.cancel()
        flag = None
        try:
            await task
        except asyncio.CancelledError:  # NOSONAR
            flag = True
        finally:
            tasks_set.discard(task)
            if log_out:
                logger.info(f"The task {task.get_name()} was cancelled {'by force' if flag else ''}")


def any2str(_x) -> str:
    return f"{_x:.10f}".rstrip('0').rstrip('.')


def f2d(_f: float) -> Decimal:
    return Decimal(str(_f))


def solve(fn, value: Decimal, x: Decimal, **kwargs) -> tuple[Decimal, str]:
    def _fn(_x):
        return abs(float(value) - fn(_x, **kwargs))
    res = minimize(_fn, x0=np.array([float(x)]), method='Nelder-Mead')
    if res.success:
        _res = f2d(res.x[0])
        n = 0
        while f2d(fn(_res, **kwargs)) - value < 0:
            _res += f2d(0.1)
            n += 1
            if n > 200:  # cycle limit check
                return O_DEC, "Number of cycles exceeded"
        return _res, f"{res.message} Number of iterations: {res.nit}, correction: +{n*0.1:.2f}"
    return O_DEC, res.message


def convert_from_minute(m: int) -> str:
    intervals = [
        (1, 3, '1m'),
        (3, 5, '3m'),
        (5, 15, '5m'),
        (15, 30, '15m'),
        (30, 60, '30m'),
        (60, 120, '1h'),
        (120, 240, '2h'),
        (240, 360, '4h'),
        (360, 480, '6h'),
        (480, 720, '8h'),
        (720, 1440, '12h'),
        (1440, 4320, '1d'),
        (4320, 10080, '3d'),
        (10080, 44640, '1w')
    ]

    for start, end, value in intervals:
        if start <= m < end:
            return value
    return '1m'  # Default case


def load_file(name: Path) -> dict:
    _res = {}
    if name.exists():
        try:
            with name.open() as state_file:
                _last_state = json.load(state_file)
        except json.JSONDecodeError as er:
            print(f"Exception on decode last state file: {er}")
        else:
            if _last_state.get('ms_start_time_ms', None):
                _res = _last_state
    return _res


def load_last_state(last_state_file) -> dict:
    res = {}
    if last_state_file.exists():
        res = load_file(last_state_file)
        if not res:
            print("Can't load last state, try load previous saved state")
            res = load_file(last_state_file.with_suffix('.prev'))
        if res:
            with last_state_file.with_suffix('.bak').open(mode='w') as outfile:
                json.dump(res, outfile, sort_keys=True, indent=4, ensure_ascii=False)
    return res


class Style:
    __slots__ = ()

    BLACK: str = '\033[30m'
    RED: str = '\033[31m'
    B_RED: str = '\033[1;31m'
    GREEN: str = '\033[32m'
    YELLOW: str = '\033[33m'
    B_YELLOW: str = "\033[33;1m"
    BLUE: str = '\033[34m'
    MAGENTA: str = '\033[35m'
    CYAN: str = '\033[36m'
    GRAY: str = '\033[37m'
    WHITE: str = '\033[0;37m'
    B_WHITE: str = '\033[1;37m'
    UNDERLINE: str = '\033[4m'
    RESET: str = '\033[0m'

    @classmethod
    def __add__(cls, b):
        return Style() + b


class Orders:
    __slots__ = ("orders_list",)

    def __init__(self):
        self.orders_list = []

    def __iter__(self):
        yield from self.orders_list

    def __len__(self):
        return len(self.orders_list)

    def append_order(self, _id: int, buy: bool, amount: Decimal, price: Decimal):
        self.orders_list.append({'id': _id, 'buy': buy, 'amount': amount, 'price': price})

    def remove(self, _id: int):
        self.orders_list[:] = [i for i in self.orders_list if i['id'] != _id]

    def find_order(self, in_orders: list, place_order_id: int):
        """
        Find equal order in_orders[] and self.orders_list[] where in_orders[].id == place_order_id
        If exist return order: Order
        """
        order = None
        for i in self.orders_list:
            if i['id'] == place_order_id:
                for k, o in enumerate(in_orders):
                    if o.buy == i['buy'] and o.amount == i['amount'] and o.price == i['price']:
                        order = in_orders[k]
                        break
            if order:
                break
        return order

    def get_by_id(self, _id: int) -> dict:
        return next((i for i in self.orders_list if i['id'] == _id), None)

    def exist(self, _id: int) -> bool:
        return any(i['id'] == _id for i in self.orders_list)

    def get(self) -> list:
        """
        Get List of Dict for orders
        :return: []
        """
        return self.orders_list

    def get_id_list(self) -> list:
        """
        Get List of orders id
        :return: []
        """
        return [i['id'] for i in self.orders_list]

    def get_first(self) -> tuple:
        """
        Get first order as tuple
        :return: (id, buy, amount, price)
        """
        return tuple(self.orders_list[0].values())

    def get_last(self) -> tuple:
        """
        Get last order as tuple
        :return: (id, buy, amount, price)
        """
        return tuple(self.orders_list[-1].values())

    def restore(self, order_list: list):
        self.orders_list.clear()
        for i in order_list:
            i_dec = {'id': i.get('id'),
                     'buy': i.get('buy'),
                     'amount': f2d(i.get('amount')),
                     'price': f2d(i.get('price'))}
            self.orders_list.append(i_dec)

    def sort(self, cycle_buy: bool):
        if cycle_buy:
            self.orders_list.sort(key=lambda x: x['price'], reverse=True)
        else:
            self.orders_list.sort(key=lambda x: x['price'], reverse=False)

    def sum_amount(self, cycle_buy: bool) -> Decimal:
        _sum = O_DEC
        for i in self.orders_list:
            _sum += i['amount'] * (i['price'] if cycle_buy else 1)
        return _sum


class PrivateTrade:
    __slots__ = (
        "amount",
        "buy",
        "is_maker",
        "id",
        "order_id",
        "price",
        "commission",
        "commission_asset",
        "timestamp"
    )

    def __init__(self, _trade: dict) -> None:
        self.amount = Decimal(_trade["qty"])
        self.buy = _trade.get('isBuyer', False)
        self.is_maker = _trade.get('isMaker', False)
        self.id = int(_trade["id"])
        self.order_id = int(_trade["orderId"])
        self.price = Decimal(_trade["price"])
        self.commission = Decimal(_trade.get('commission', "0"))
        self.commission_asset = _trade.get('commissionAsset', "")
        self.timestamp = int(_trade["time"])

    def __call__(self):
        return self


class OrderUpdate:
    __slots__ = ("original_order", "resulting_trades", "status", "timestamp", "updated_order")

    class Status(Enum):
        """
        Update status defining what happened to the order since the last update.
        """
        FILLED = 0
        ADAPTED = 1
        CANCELED = 2
        NO_CHANGE = 3
        REAPPEARED = 4
        DISAPPEARED = 5
        OTHER_CHANGE = 6
        PARTIALLY_FILLED = 7
        ADAPTED_AND_FILLED = 8

    ADAPTED = Status.ADAPTED
    ADAPTED_AND_FILLED = Status.ADAPTED_AND_FILLED
    CANCELED = Status.CANCELED
    DISAPPEARED = Status.DISAPPEARED
    FILLED = Status.FILLED
    NO_CHANGE = Status.NO_CHANGE
    OTHER_CHANGE = Status.OTHER_CHANGE
    PARTIALLY_FILLED = Status.PARTIALLY_FILLED
    REAPPEARED = Status.REAPPEARED

    def __init__(self, event: dict, trades: list) -> None:

        class OriginalOrder:
            __slots__ = ("id",)

            def __init__(self, _event: dict):
                self.id = _event['order_id']

        self.original_order = OriginalOrder(event)
        self.resulting_trades = []
        for trade in trades:
            if trade.order_id == event['order_id']:
                self.resulting_trades.append(trade)
        if event['order_status'] == 'FILLED':
            self.status = OrderUpdate.FILLED
        elif event['order_status'] == 'PARTIALLY_FILLED':
            self.status = OrderUpdate.PARTIALLY_FILLED
        elif event['order_status'] == 'CANCELED':
            self.status = OrderUpdate.CANCELED
        else:
            self.status = OrderUpdate.OTHER_CHANGE
        self.timestamp = event['transaction_time']
        self.updated_order = None

    def __call__(self):
        return self


class Order:
    __slots__ = ("amount", "buy", "id", "order_type", "price", "received_amount", "remaining_amount", "timestamp")

    def __init__(self, order: dict):
        self.amount = Decimal(order['origQty'])
        self.buy = order['side'] == 'BUY'
        self.id = int(order['orderId'])
        self.order_type = order['type']
        self.received_amount = Decimal(order['executedQty'])
        cummulative_quote_qty = order.get('cummulativeQuoteQty')
        if self.received_amount > 0 and cummulative_quote_qty:
            self.price = Decimal(cummulative_quote_qty) / self.received_amount
        else:
            self.price = Decimal(order['price'])
        self.remaining_amount = self.amount - self.received_amount
        self.timestamp = int(order.get('transactTime', order.get('time', time.time())))

    def __call__(self):
        return self


class Candle:
    __slots__ = ("min_time", "open", "high", "low", "close", "volume", "max_time", "trade_number", "vwap")

    def __init__(self, _candle: list):
        self.min_time = int(_candle[0])
        self.open = float(_candle[1])
        self.high = float(_candle[2])
        self.low = float(_candle[3])
        self.close = float(_candle[4])
        self.volume = float(_candle[5])
        self.max_time = int(_candle[6])
        self.trade_number = int(_candle[8])
        self.vwap = (float(_candle[7]) / self.volume) if self.volume else self.close

    def __call__(self):
        return self


class TradingCapabilityManager:
    __slots__ = (
        "base_asset_precision",
        "quote_asset_precision",
        "min_qty",
        "max_qty",
        "step_size",
        "min_notional",
        "tick_size",
        "multiplier_up",
        "multiplier_down",
        "min_price",
        "max_price",
    )

    def __init__(self, _exchange_info_symbol):
        self.base_asset_precision = int(_exchange_info_symbol.get('baseAssetPrecision'))
        self.quote_asset_precision = int(_exchange_info_symbol.get('quoteAssetPrecision'))
        self.min_qty = Decimal(_exchange_info_symbol['filters']['lotSize']['minQty'])
        self.max_qty = Decimal(_exchange_info_symbol['filters']['lotSize']['maxQty'])
        self.step_size = Decimal(_exchange_info_symbol['filters']['lotSize']['stepSize'].rstrip('0'))
        self.min_notional = (
                Decimal(_exchange_info_symbol['filters'].get('notional', {}).get('minNotional', '0'))
                or Decimal(_exchange_info_symbol['filters'].get('minNotional', {}).get('minNotional', '0'))
        )
        self.tick_size = Decimal(_exchange_info_symbol['filters']['priceFilter']['tickSize'].rstrip('0'))
        self.min_price = Decimal(_exchange_info_symbol['filters']['priceFilter']['minPrice'])
        self.max_price = Decimal(_exchange_info_symbol['filters']['priceFilter']['maxPrice'])
        self.multiplier_up = Decimal(_exchange_info_symbol['filters']['percentPrice']['multiplierUp'])
        self.multiplier_down = Decimal(_exchange_info_symbol['filters']['percentPrice']['multiplierDown'])

    def __call__(self):
        return self

    def round_amount(self, unrounded_amount: Decimal, rounding_type: str) -> Decimal:
        return unrounded_amount.quantize(self.step_size, rounding=rounding_type)

    def round_price(self, unrounded_price: Decimal, rounding_type: str) -> Decimal:
        return unrounded_price.quantize(self.tick_size, rounding=rounding_type)

    def get_min_sell_amount(self, price: Decimal) -> Decimal:
        return max(self.min_qty, self.round_amount(self.min_notional / price, ROUND_CEILING))

    def get_max_sell_amount(self, _unused_price: Decimal) -> Decimal:
        """
        Returns the maximally possible sell amount that can be placed at a given price.
        """
        return self.max_qty

    def get_min_buy_amount(self, price: Decimal) -> Decimal:
        return max(self.min_qty, self.round_amount(self.min_notional / price, ROUND_CEILING))

    def get_minimal_price_change(self) -> Decimal:
        return self.tick_size

    def get_minimal_amount_change(self) -> Decimal:
        """
        Get the minimal amount change that is possible to use on the exchange.
        """
        return self.step_size

    def get_max_sell_price(self, avg_price: Decimal) -> Decimal:
        return self.round_price(avg_price * self.multiplier_up, ROUND_FLOOR)

    def get_max_price(self) -> Decimal:
        return self.max_price

    def get_min_buy_price(self, avg_price: Decimal) -> Decimal:
        return self.round_price(avg_price * self.multiplier_down, ROUND_CEILING)

    def get_min_price(self) -> Decimal:
        return self.min_price


class Ticker:
    __slots__ = ("last_day_price", "last_price", "timestamp")

    def __init__(self, _ticker):
        self.last_day_price = Decimal(_ticker['openPrice'])
        self.last_price = Decimal(_ticker['lastPrice'])
        self.timestamp = int(_ticker['closeTime'])

    def __call__(self):
        return self


class FundsEntry:
    __slots__ = ("available", "reserved", "total_for_currency")

    def __init__(self, _funds):
        self.available = Decimal(_funds['free'])
        self.reserved = Decimal(_funds['locked'])
        self.total_for_currency = self.available + self.reserved

    def __call__(self):
        return self


class OrderBook:
    __slots__ = ("asks", "bids")

    """
    order_book.bids[0].price
    order_book.asks[0].amount
    """

    def __init__(self, _order_book, _tcm=None) -> None:
        class _OrderBookRow:
            __slots__ = ("price", "amount")

            def __init__(self, _order, _tcm=_tcm) -> None:
                self.price = Decimal(_order[0])
                self.amount = Decimal(_order[1])
                if _tcm:
                    self.price = _tcm.round_price(self.price, ROUND_HALF_EVEN)
                    self.amount = _tcm.round_amount(self.amount, ROUND_HALF_EVEN)

        self.asks = []
        self.bids = []
        self.asks.extend(_OrderBookRow(v) for v in _order_book['asks'])
        self.bids.extend(_OrderBookRow(v) for v in _order_book['bids'])

    def __call__(self):
        return self


class Klines:
    klines_series = {}
    klines_lim = int()

    def __init__(self, _interval):
        self.interval = _interval
        self.kline = []
        self.klines_series[_interval] = self.kline

    def refresh(self, _candle):
        candle = Candle(_candle)
        new_time = candle.min_time
        last_time = self.kline[-1].min_time if self.kline else 0
        if new_time >= last_time:
            if new_time == last_time:
                self.kline[-1] = candle
            else:
                self.kline.append(candle)
                if len(self.kline) > self.klines_lim:
                    del self.kline[0]
            self.klines_series[self.interval] = self.kline

    @classmethod
    def get_kline(cls, _interval) -> list:
        return cls.klines_series.get(_interval, [])



"""
martin-binance strategy parameters
"""
__author__ = "Jerry Fedorenko"
__copyright__ = "Copyright © 2021-2025 Jerry Fedorenko aka VM"
__license__ = "MIT"
__version__ = "3.0.20"
__maintainer__ = "Jerry Fedorenko"
__contact__ = "https://github.com/DogsTailFarmer"

import logging
from decimal import Decimal
from pathlib import Path

__all__ = [
    'SYMBOL', 'EXCHANGE', 'ID_EXCHANGE', 'FEE_MAKER', 'FEE_TAKER', 'FEE_FIRST', 'FEE_SECOND', 'FEE_BNB',
    'SAVE_ASSET', 'GRID_MAX_COUNT', 'START_ON_BUY', 'AMOUNT_FIRST', 'USE_ALL_FUND', 'AMOUNT_SECOND',
    'PRICE_SHIFT', 'ROUND_BASE', 'ROUND_QUOTE', 'PROFIT', 'PROFIT_MAX', 'OVER_PRICE', 'ORDER_Q',
    'MARTIN', 'SHIFT_GRID_DELAY', 'GRID_UPDATE_INTERVAL', 'STATUS_DELAY', 'GRID_ONLY', 'LOG_LEVEL',
    'HOLD_TP_ORDER_TIMEOUT', 'COLLECT_ASSETS', 'GRID_ONLY_DELAY', 'TP_REFRESH', 'ADAPTIVE_TRADE_CONDITION',
    'BB_CANDLE_SIZE_IN_MINUTES', 'BB_NUMBER_OF_CANDLES', 'KBB', 'LINEAR_GRID_K', 'ADX_CANDLE_SIZE_IN_MINUTES',
    'ADX_NUMBER_OF_CANDLES', 'ADX_PERIOD', 'ADX_THRESHOLD', 'ADX_PRICE_THRESHOLD', 'REVERSE', 'REVERSE_TARGET_AMOUNT',
    'REVERSE_INIT_AMOUNT', 'REVERSE_STOP', 'HEAD_VERSION', 'LOAD_LAST_STATE', 'LAST_STATE_FILE', 'VPS_NAME', 'PARAMS',
    'TELEGRAM_CONFIG', 'MODE', 'XTIME', 'SAVE_DS', 'SAVE_PERIOD', 'LOGGING', 'SELF_OPTIMIZATION', 'N_TRIALS',
    'SESSION_RESULT'
]

SYMBOL = str()
EXCHANGE = ()
# Exchange setup
ID_EXCHANGE = int()
FEE_MAKER = Decimal()
FEE_TAKER = Decimal()
FEE_FIRST = False
FEE_SECOND = False
FEE_BNB = {
    'id_exchange': 0,
    'symbol': 'BNB/USDT',
    'email': str(),
    'target_amount': '0',
    'tranche_volume': '0'
}
SAVE_ASSET = True
GRID_MAX_COUNT = int()
# Trade parameter
START_ON_BUY = bool()
AMOUNT_FIRST = Decimal()
USE_ALL_FUND = bool()
AMOUNT_SECOND = Decimal()
PRICE_SHIFT = Decimal()
# Round pattern
ROUND_BASE = str()
ROUND_QUOTE = str()
#
PROFIT = Decimal()
PROFIT_MAX = Decimal()
OVER_PRICE = Decimal()
ORDER_Q = int()
MARTIN = Decimal()
SHIFT_GRID_DELAY = int()
GRID_UPDATE_INTERVAL = 60 * 60  # sec between grid update in Reverse cycle
# Other
STATUS_DELAY = int()
GRID_ONLY = bool()
LOG_LEVEL = logging.DEBUG  # Default level for console output
HOLD_TP_ORDER_TIMEOUT = 30
COLLECT_ASSETS = bool()
GRID_ONLY_DELAY = 150  # sec delay before try restart GRID_ONLY cycle
TP_REFRESH = 60 * 10  # sec between TP refresh
#
ADAPTIVE_TRADE_CONDITION = bool()
BB_CANDLE_SIZE_IN_MINUTES = int()
BB_NUMBER_OF_CANDLES = int()
KBB = float()
#
LINEAR_GRID_K = int()
#
ADX_CANDLE_SIZE_IN_MINUTES = int()
ADX_NUMBER_OF_CANDLES = int()
ADX_PERIOD = int()
ADX_THRESHOLD = int()
ADX_PRICE_THRESHOLD = float()
# Reverse cycle
REVERSE = bool()
REVERSE_TARGET_AMOUNT = Decimal()
REVERSE_INIT_AMOUNT = Decimal()
REVERSE_STOP = bool()
# Config variables
HEAD_VERSION = str()
LOAD_LAST_STATE = int()
# Path and files name
LAST_STATE_FILE: Path
VPS_NAME = str()
PARAMS: Path
# Telegram
TLG_SERVICE = False
TELEGRAM_CONFIG = {}
TLG_INLINE = False
# Backtesting
MODE = 'T'  # 'T' - Trade, 'TC' - Trade and Collect, 'S' - Simulate
XTIME = 1000  # Time accelerator
SAVE_DS = False  # Save session result data (ticker, orders) for compare
SAVE_PERIOD = 1 * 60 * 60  # sec, timetable for save data portion
LOGGING = True
SELF_OPTIMIZATION = True  # Cyclic self-optimization of parameters, together with MODE == 'TC'
N_TRIALS = 250  # Number of optimization cycles for optuna study
SESSION_RESULT = {}
